/*
**                   tc2cfr v1.5 - August 02, 2006
**
**   This program turns an avi file (with each frame present once) into
**   a new avi file w/ null frames with the designated framerate based
**   off a v1 or v2 timecode file specifying the display length for
**   each frame.
**
**   Copyright (C) 2005-2006 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <stdio.h>
#include <time.h>
#include <malloc.h>

#define VERSION "v1.5"

typedef DWORD FOURCC;

typedef struct
{
	FOURCC riff;
	DWORD fileSize;
	DWORD fileType;
} RIFFHEADER;

class CHUNKHEADER
{
public:
	FOURCC ckID;
	DWORD ckSize;
	DWORD ckDataSize;
	void *ckData;
	CHUNKHEADER::CHUNKHEADER() : ckID(0), ckSize(0), ckData(0), ckDataSize(0) {};
	CHUNKHEADER::~CHUNKHEADER() { if (ckData) free(ckData); }
};

class LISTHEADER
{
public:
	FOURCC lstID;
	DWORD lstSize;
	FOURCC lstType;
	CHUNKHEADER *chkHd;
	LISTHEADER::LISTHEADER() : lstID(0), lstSize(0), lstType(0), chkHd(0) {};
	LISTHEADER::~LISTHEADER() { if (chkHd) delete chkHd; }
};

class TBUF
{
public:
	DWORD size;
	void *data;
	TBUF::TBUF() : data(0), size(0) {};
	TBUF::~TBUF() { if (data) free(data); }
};
TBUF tbuf;

typedef struct 
{
	FOURCC fccType;
	FOURCC fccHandler;
	DWORD  dwFlags;
	WORD   wPriority;
	WORD   wLanguage;
	DWORD  dwInitialFrames;
	DWORD  dwScale;
	DWORD  dwRate;
	DWORD  dwStart;
	DWORD  dwLength;
	DWORD  dwSuggestedBufferSize;
	DWORD  dwQuality;
	DWORD  dwSampleSize;
	struct {
		short int left;
		short int top;
		short int right;
		short int bottom;
	}  rcFrame;
} AVISTREAMHEADER;

typedef struct 
{
	DWORD  dwMicroSecPerFrame;
	DWORD  dwMaxBytesPerSec;
	DWORD  dwPaddingGranularity;
	DWORD  dwFlags;
	DWORD  dwTotalFrames;
	DWORD  dwInitialFrames;
	DWORD  dwStreams;
	DWORD  dwSuggestedBufferSize;
	DWORD  dwWidth;
	DWORD  dwHeight;
	DWORD  dwReserved[4];
} AVIMAINHEADER;

typedef struct 
{
	DWORD   dwGrandFrames;
	DWORD   dwFuture[61];
} AVIEXTHEADER;

typedef struct
{
	DWORD   dwChunkId;
	DWORD   dwFlags;
	DWORD   dwOffset;
	DWORD   dwSize;
} AVIOLDINDEX;

DWORD MAKE_FOURCC(char* fourCC)
{
	return (fourCC[0]|(fourCC[1]<<8)|(fourCC[2]<<16)|(fourCC[3]<<24));
}

bool checkRiffHeader(RIFFHEADER *rh)
{
	if (rh->riff != MAKE_FOURCC("RIFF")) return false;
	if (rh->fileType != MAKE_FOURCC("AVI ")) return false;
	return true;
}

bool scanForNextList(FILE *f, FILE *o, LISTHEADER *lstH, char *fourCC, bool force,
					 DWORD nsize, bool &fmovi)
{
repeat:
	if (!fread(&lstH->lstID,1,2,f)) return false;
	while(1)
	{
		if (!fread(((unsigned char*)&lstH->lstID)+2,1,2,f)) return false;
		if (lstH->lstID == 0x5453494C) break;
		if (!fwrite(&lstH->lstID,1,2,o)) return false;
		lstH->lstID = lstH->lstID>>16;
	}
	if (!fread(&lstH->lstSize,4,1,f)) return false;
	if (!fread(&lstH->lstType,4,1,f)) return false;
	if (lstH->lstType != MAKE_FOURCC(fourCC)) 
	{
		if (force) return false;
		if (MAKE_FOURCC(fourCC) == MAKE_FOURCC("odml") &&
			lstH->lstType == MAKE_FOURCC("movi"))
		{
			if (fseek(f,-12,SEEK_CUR)) return false;
			fmovi = true;
			return false;
		}
		if (!fwrite(&lstH->lstID,4,1,o)) return false;
		if (!fwrite(&lstH->lstSize,4,1,o)) return false;
		if (!fwrite(&lstH->lstType,4,1,o)) return false;
		DWORD fsize = lstH->lstSize-4-(lstH->lstSize&1);
		if (tbuf.size <= fsize+12)
		{
			if (tbuf.data) free(tbuf.data);
			tbuf.data = malloc(fsize+12);
			if (!tbuf.data) return false;
			tbuf.size = fsize+12;
		}
		if (!fread(tbuf.data,fsize,1,f)) return false;
		if (!fwrite(tbuf.data,fsize,1,o)) return false;
		goto repeat;
	}
	if (!fwrite(&lstH->lstID,4,1,o)) return false;
	lstH->lstSize += nsize;
	if (!fwrite(&lstH->lstSize,4,1,o)) return false;
	if (!fwrite(&lstH->lstType,4,1,o)) return false;
	return true;
}

bool readChunk(FILE *f, FILE *o, char *fourCC, DWORD size, LISTHEADER *lh)
{
	if (lh->chkHd) delete lh->chkHd;
	lh->chkHd = new CHUNKHEADER();
	if (!lh->chkHd) return false;
	if (!fread(&lh->chkHd->ckID,4,1,f)) return false;
	if (!fwrite(&lh->chkHd->ckID,4,1,o)) return false;
	if (lh->chkHd->ckID != MAKE_FOURCC(fourCC)) return false;
	if (!fread(&lh->chkHd->ckSize,4,1,f)) return false;
	if (!fwrite(&lh->chkHd->ckSize,4,1,o)) return false;
	if (lh->chkHd->ckSize != size) return false;
	if (lh->chkHd->ckData) free(lh->chkHd->ckData);
	lh->chkHd->ckDataSize = lh->chkHd->ckSize+(lh->chkHd->ckSize&1);
	lh->chkHd->ckData = malloc(lh->chkHd->ckDataSize);
	if (!lh->chkHd->ckData) return false;
	if (!fread(lh->chkHd->ckData, lh->chkHd->ckSize, 1, f)) return false;
	return true;
}

bool readNextChunk(FILE *f, FILE *o, CHUNKHEADER *chnk, char *fourCC, bool read, DWORD nsize)
{
repeat:
	if (!fread(&chnk->ckID,4,1,f)) return false;
	if (!fwrite(&chnk->ckID,4,1,o)) return false;
	if (!fread(&chnk->ckSize,4,1,f)) return false;
	if (nsize == 0) { if (!fwrite(&chnk->ckSize,4,1,o)) return false; }
	else 
	{
		chnk->ckSize += nsize;
		if (!fwrite(&chnk->ckSize,4,1,o)) return false;
		chnk->ckSize -= nsize;
	}
	DWORD pad_size = chnk->ckSize+(chnk->ckSize&1);
	if (chnk->ckID == MAKE_FOURCC(fourCC) ||
		(MAKE_FOURCC(fourCC) == MAKE_FOURCC("00dc") && chnk->ckID == MAKE_FOURCC("00db")) ||
		(MAKE_FOURCC(fourCC) == MAKE_FOURCC("00db") && chnk->ckID == MAKE_FOURCC("00dc")))
	{
		if (!read) return true;
		if (chnk->ckDataSize <= pad_size+12) 
		{
			if (chnk->ckData) free(chnk->ckData);
			chnk->ckData = malloc(pad_size+12);
			if (!chnk->ckData) return false;
			chnk->ckDataSize = pad_size+12;
		}
		if (!fread(chnk->ckData,pad_size,1,f)) return false;
		if (!fwrite(chnk->ckData,pad_size,1,o)) return false;
		return true;
	}
	else
	{
		if (tbuf.size <= pad_size+12)
		{
			if (tbuf.data) free(tbuf.data);
			tbuf.data = malloc(pad_size+12);
			if (!tbuf.data) return false;
			tbuf.size = pad_size+12;
		}
		if (!fread(tbuf.data,pad_size,1,f)) return false;
		if (!fwrite(tbuf.data,pad_size,1,o)) return false;
		goto repeat;
	}
	return false;
}

bool readNextIndexChunk(FILE *f, AVIOLDINDEX *index)
{
	if (!fread(index,1,16,f)) return false;
	return true;
}

void closeFiles(FILE *f, FILE *o, FILE *tcf, DWORD *repCount, bool ul, char *oname, 
				char *fbuf, char *obuf)
{
	if (f) fclose(f);
	if (o) 
	{
		fclose(o);
		if (ul && oname) unlink(oname);
	}
	if (tcf) fclose(tcf);
	if (repCount) free(repCount);
	if (fbuf) _aligned_free(fbuf);
	if (obuf) _aligned_free(obuf);
}

bool writeEmptyFrames(FILE *o, int count, CHUNKHEADER *chnk)
{
	char buf[8] = { '0', '0', 'd', 'c', 0x00, 0x00, 0x00, 0x00 };
	if (chnk->ckID == MAKE_FOURCC("00db")) buf[3] = 'b';
	for (int i=0; i<count; ++i)
	{
		if (!fwrite(buf, 8, 1, o)) return false;
	}
	return true;
}

bool writeNewIndexEntries(FILE *o, int count, AVIOLDINDEX *index, DWORD offset)
{
	index->dwSize = 0;
	index->dwFlags &= ~(0x00000010);
	index->dwOffset = offset;
	for (int i=0; i<count; ++i)
	{
		if (!fwrite(index,sizeof(AVIOLDINDEX),1,o)) return false;
		index->dwOffset += 8;
	}
	return true;
}

void findRateAndScale(double fps, DWORD &rate, DWORD &scale)
{
	DWORD d = 1;
	while (d < 10000000 && fps < 10000000.0)
	{
		d *= 10;
		fps *= 10.0;
	}
	DWORD x = int(fps+0.5);
	DWORD y = d, t;
	while (y) {
		t = x%y; 
		x = y;
		y = t;
	}
	rate = int(fps+0.5)/x;
	scale = d/x;
}

int checkValidFPS(double fps, int **fps_table, int ecount)
{
	int ifps = int(fps*1000.0+0.5);
	for (int j=0; j<ecount; ++j)
	{
		for (int k=0; k<3; ++k)
		{
			if (fps_table[j][k] == ifps)
			{
				if (j != 10000) return j;
				else return 100000;
			}
		}
	}
	return -1;
}

void createFPSTable(double fps, int **fps_table, bool edec)
{
	int j;
	for (j=0; j<10000; ++j)
	{
		double tfps = double(j+1)*double(1000.0)/fps;
		fps_table[j][0] = int((1000000.0/double(int(tfps))) + 0.5);
		fps_table[j][1] = int((1000000.0/tfps) + 0.5);
		fps_table[j][2] = int((1000000.0/double(int(tfps+1.0))) + 0.5);
	}
	for (j=0; j<9999; ++j)
	{
		if (fps_table[j][0] == fps_table[j+1][0]) break;
		if (fps_table[j][1] == fps_table[j+1][1]) break;
		if (fps_table[j][2] == fps_table[j+1][2]) break;
	}
	printf("Descernable number of framerates: %d (min = %.6f fps).\n", j, fps/double(j));
	for (; j<10000; ++j)
	{
		fps_table[j][0] = -20;
		fps_table[j][1] = -20;
		fps_table[j][2] = -20;
	}
	if (edec)
	{
		fps_table[10000][0] = 18182;
		fps_table[10000][1] = 17982;
		fps_table[10000][2] = 17857;
	}
}

void destroy_fps_table(int **&fps_table, int ecount)
{
	if (!fps_table) return;
	for (int i=0; i<ecount; ++i)
	{
		if (fps_table[i]) free(fps_table[i]);
	}
	free(fps_table);
	fps_table = NULL;
}

bool init_fps_table(double fps, int **&fps_table, bool edec)
{
	int ecount = edec ? 10001 : 10000;
	fps_table = (int **)malloc(ecount*sizeof(int*));
	if (!fps_table)
	{
		printf("malloc failure (fps_table 1).\n");
		return false;
	}
	memset(fps_table,0,ecount*sizeof(int*));
	for (int k=0; k<ecount; ++k)
	{
		fps_table[k] = (int *)malloc(3*sizeof(int));
		if (!fps_table[k])
		{
			printf("malloc failure (fps_table 2).\n");
			destroy_fps_table(fps_table, ecount);
			return false;
		}
	}
	createFPSTable(fps, fps_table, edec);
	return true;
}

bool parseV1TCFile(FILE *tcf, int fcount, DWORD *repCount, int **fps_table, int ecount)
{
	char buf[512];
	if (!fgets(buf,80,tcf))
	{
		printf("Error reading from v1 timecode file.\n");
		return false;
	}
	int i, f1, f2;
	double defaultfps, fps;
	bool tcl = false;
	if (strncmp(buf,"Assume ", 7))
	{
		printf("Invalid v1 timecode file (incorrect Assume fps line).\n");
		return false;
	}
	sscanf(buf,"Assume %lf", &defaultfps);
	int temp = checkValidFPS(defaultfps, fps_table, ecount);
	if (temp < 0 || temp == 100000)
	{
		printf("Invalid default fps for v1 timecode file (%f).\n", defaultfps);
		return false;
	}
	for (i=0; i<fcount; ++i) repCount[i] = temp;
	while(fgets(buf,80,tcf))
	{
		if (buf[0] == '#' || buf[0] == '\n' || buf[0] == ' ' || 
			buf[0] == '\r' || buf[0] == 0) continue;
		sscanf(buf,"%d,%d,%lf", &f1, &f2, &fps);
		temp = checkValidFPS(fps, fps_table, ecount);
		if (temp < 0)
		{
			fps = double(1000.0)/double(int(double(1000.0)/fps+0.5));
			temp = checkValidFPS(fps, fps_table, ecount);
			if (temp < 0)
			{
				printf("Invalid fps detected (%f).\n", fps);
				return false;
			}
			else if (!tcl)
			{
				tcl = true;
				printf("Switching to lenient timecode validation for v1 timecode file.\n");
			}
		}
		if (f1 < 0 || f2 < 0 || f1 >= fcount || f2 >= fcount || f1 > f2)
		{
			printf("Invalid frame range detected in v1 timecode file (%d-%d).\n", f1, f2);
			return false;
		}
		if (temp != 100000)
		{
			for (i=f1; i<=f2; ++i) 
				repCount[i] = temp;
		}
		else
		{
			if ((f2-f1+1)%3 != 0)
			{
				printf("17.982 frame rate can only appear in ranges that are mod 3.\n");
				return false;
			}
			for (i=f1; i<=f2; i+=3)
			{
				repCount[i+0] = 6;
				repCount[i+1] = 6;
				repCount[i+2] = 5;
			}
		}
	}
	return true;
}

bool parseV2TCFile(FILE *tcf, int fcount, DWORD *repCount, int **fps_table, int ecount)
{
	char buf[512];
	bool first = true;
	int i = 0, temp, ctrack = 0, divf;
	double timeo, time = -20.0, fps;
	while(fgets(buf,80,tcf))
	{
		if (buf[0] == '#' || buf[0] == '\n' || buf[0] == ' ' || 
			buf[0] == '\r' || buf[0] == 0) continue;
		timeo = time;
		sscanf(buf,"%lf", &time);
		if (timeo > time || time < 0.0)
		{
			printf("Out of order timecodes detected in v2 file.\n");
			return false;
		}
		if (first) { first = false; continue; }
		divf = int(time-timeo);
		if (divf == 0)
		{
			printf("Invalid framerate detected in v2 timecode file.\n");
			return false;
		}
		fps = double(1000.0)/double(divf);
		temp = checkValidFPS(fps, fps_table, ecount);
		if (temp < 0)
		{
			printf("Invalid fps detected in v2 timecode file (%f).\n", fps);
			return false;
		}
		else if (temp != 100000) 
		{
			if (ctrack)
			{
				printf("17.982 fps can only occur in multiples of three frames.\n");
				return false;
			}
			repCount[i++] = temp;
		}
		else
		{
			++ctrack;
			if (ctrack != 0 && (ctrack%3) == 0)
			{
				repCount[i+0] = 6;
				repCount[i+1] = 6;
				repCount[i+2] = 5;
				i += 3;
				ctrack -= 3;
			}
		}
	}
	if (ctrack == 2)
	{
		repCount[i+0] = 6;
		repCount[i+1] = 6;
		repCount[i+2] = 5;
		i += 3;
	}
	else if (ctrack)
	{
		printf("17.982 fps can only occur in multiples of three frames.\n");
		return false;
	}
	else repCount[i++] = temp;
	if (i != fcount)
	{
		printf("Not all frames have durations in v2 timecode file.\n");
		return false;
	}
	return true;
}

bool parseTimecodeFile(FILE *tcf, int fcount, DWORD *&repCount, DWORD &newFrameCount, 
					   double fps, double ifps)
{
	char buf[512];
	repCount = (DWORD*)malloc(fcount*sizeof(DWORD));
	if (!repCount)
	{
		printf("Malloc failure (repCount).\n");
		return false;
	}
	if (!tcf && ifps <= 0.0)
		return false;
	if (tcf)
	{
		if (!fgets(buf,80,tcf)) 
		{
			printf("Error reading from timecode file (0).\n");
			return false;
		}
	}
	int **fps_table = NULL;
	bool edec = (fps >= 119.880000 && fps <= 119.880120) ? true : false;
	if (tcf)
	{
		if (strncmp(buf,"# timecode format v1", 20))
		{
			if (strncmp(buf,"# timecode format v2", 20))
			{
				printf("Invalid timecode file (incorrect first line).\n");
				return false;
			}
			if (!init_fps_table(fps, fps_table, edec)) return false;
			printf("v2 timecode file type detected.\n");
			if (!parseV2TCFile(tcf, fcount, repCount, fps_table, edec ? 10001 : 10000)) 
			{
				destroy_fps_table(fps_table, edec ? 10001 : 10000);
				return false;
			}
		}
		else 
		{
			if (!init_fps_table(fps, fps_table, edec)) return false;
			printf("v1 timecode file type detected.\n");
			if (!parseV1TCFile(tcf, fcount, repCount, fps_table, edec ? 10001 : 10000)) 
			{
				destroy_fps_table(fps_table, edec ? 10001 : 10000);
				return false;
			}
		}
	}
	else
	{
		if (!init_fps_table(fps, fps_table, edec)) return false;
		int temp = checkValidFPS(ifps,fps_table, edec ? 10001 : 10000);
		if (temp < 0 || temp == 100000)
		{
			double ifpst = double(1000.0)/double(int(double(1000.0)/ifps+0.5));
			temp = checkValidFPS(ifpst, fps_table, edec ? 10001 : 10000);
			if (temp < 0 || temp == 100000)
			{
				printf("Invalid input fps detected (%f).\n", ifps);
				destroy_fps_table(fps_table, edec ? 10001 : 10000);
				return false;
			}
			else printf("Switching to lenient timecode validation for input fps.\n");
		}
		for (int i=0; i<fcount; ++i)
			repCount[i] = temp;
	}
	destroy_fps_table(fps_table, edec ? 10001 : 10000);
	newFrameCount = 0;
	for (int i=0; i<fcount; ++i) 
		newFrameCount += repCount[i]+1;
	return true;
}

int main(int argc, char *argv[])
{
	DWORD frameCount, newFrameCount;
	DWORD *repCount = 0;
	clock_t start, finish;
	bool fmovi = false, frac = false;
	char *fbuf = 0, *obuf = 0;
	double fps;
	unsigned int num, den;

	printf("\ntc2cfr %s by tritical.\n", VERSION);

	/* Check for valid syntax. */
	if (argc != 5)
	{
		printf("Invalid syntax used.\n");
		printf("Usage:    tc2cfr out_fps input_avi_file timecode_file output_avi_file\n");
		printf("Example:  tc2cfr 119.880120 c:\\video.avi c:\\timecodes.txt c:\\video-120.avi\n");
		printf("Example:  tc2cfr 120000/1001 c:\\video.avi c:\\timecodes.txt c:\\video-120.avi\n");
		return 1;
	}
	start = clock();

	/* Check for valid fps setting. */
	char *p = argv[1];
	while (*p != 0 && *p != '/') p++;
	if (*p == 0) 
	{
		fps = atof(argv[1]);
		if (fps <= 0.0)
		{
			printf("Invalid out_fps setting (double).\n");
			return 2;
		}
	}
	else
	{
		if (sscanf(argv[1],"%u/%u", &num, &den) != 2)
		{
			printf("Invalid out_fps setting (fraction).\n");
			return 3;
		}
		if (num == 0 || den == 0)
		{
			printf("Invalid out_fps setting (fraction).\n");
			return 3;
		}
		fps = double(num)/double(den);
		frac = true;
	}

	/* Open the specified avi file. */
	FILE *f = fopen(argv[2],"rb");
	if (!f)
	{
		printf("Error opening input file.\n");
		return 4;
	}
	fbuf = (char*)_aligned_malloc(65536*sizeof(char),32);
	if (!fbuf)
	{
		printf("Error allocating file buffer for input avi file.\n");
		fclose(f);
		return 5;
	}
	if (setvbuf(f, fbuf, _IOFBF, 65536))
	{
		printf("Error setting input avi file buffer.\n");
		closeFiles(f, NULL, NULL, repCount, false, NULL, fbuf, obuf);
		return 6;
	}

	/* Open the timecode file. */
	double ifps = -20.0;
	FILE *tcf = fopen(argv[3],"r");
	if (!tcf)
	{
		/* See if it's an fps value */
		bool gv = true;
		char *p = argv[3];
		while (*p != 0 && *p != '/') p++;
		if (*p == 0) 
		{
			if (sscanf(argv[3],"%lf", &ifps) != 1)
				gv = false;
			if (ifps <= 0.0)
				gv = false;
		}
		else
		{
			unsigned int numi, deni;
			if (sscanf(argv[3],"%u/%u", &numi, &deni) != 2)
				gv = false;
			if (numi == 0 || deni == 0)
				gv = false;
			else ifps = double(numi)/double(deni);
		}
		if (!gv)
		{
			printf("Error opening timecode file.\n");
			closeFiles(f, NULL, NULL, repCount, false, NULL, fbuf, obuf);
			return 7;
		}
	}

	/* Read 12 byte RIFF header and check that it is an AVI file. */
	RIFFHEADER rh;
	if (!fread(&rh,1,12,f))
	{
		printf("Error reading beginning of file.\n");
		closeFiles(f, NULL, tcf, repCount, true, NULL, fbuf, obuf);
		return 8;
	}
	if(!checkRiffHeader(&rh)) 
	{
		printf("Invalid RIFF header (not an avi file).\n");
		closeFiles(f, NULL, tcf, repCount, true, NULL, fbuf, obuf);
		return 9;
	}

	/* Open the output file. */
	FILE *o = fopen(argv[4],"wb");
	if (!o)
	{
		printf("Error opening output file.\n");
		closeFiles(f, NULL, tcf, repCount, true, NULL, fbuf, obuf);
		return 10;
	}
	obuf = (char*)_aligned_malloc(65536*sizeof(char),32);
	if (!obuf)
	{
		printf("Error allocating file buffer for output avi file.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 11;
	}
	if (setvbuf(o, obuf, _IOFBF, 65536))
	{
		printf("Error setting output avi file buffer.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 12;
	}
	if (!fwrite(&rh,1,12,o))
	{
		printf("Error writing beginning of output file.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 13;
	}

	/* Scan until first "LIST" fourcc.  The first one should be "hdrl" type. */
	LISTHEADER hdrl;
	if (!scanForNextList(f, o, &hdrl, "hdrl", true, 0, fmovi))
	{
		printf("First LIST is invalid.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 14;
	}

	/* Next should be the "avih" chunk with the main AVI Header. */
	if (!readChunk(f, o, "avih", sizeof(AVIMAINHEADER), &hdrl))
	{
		printf("Error reading avih CHUNK.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 15;
	}
	AVIMAINHEADER* am = (AVIMAINHEADER*)hdrl.chkHd->ckData;

	/* Make sure that the file has only one stream present. */
	if (am->dwStreams != 1)
	{
		printf("File has more than one stream present.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 16;
	}

	/* Make sure that the file has an "idx1" index. */
	if (!(am->dwFlags&0x00000010))
	{
		printf("File has no \"idx1\" index (dwFlags).\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 17;
	}

	/* Make sure that AVIF_MUSTUSEINDEX is not set. */
	if (am->dwFlags&0x00000020)
	{
		printf("Cannot use a file with AVIF_MUSTUSEINDEX flag set.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 18;
	}

	/* Modify needed fields in AVIMAINHEADER structure. */
	frameCount = am->dwTotalFrames;
	/* Parse the timecode file and get the new framecount. */
	if (!parseTimecodeFile(tcf, frameCount, repCount, newFrameCount, fps, ifps))
	{
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 19;
	}
	printf("Timecode file parsed successfully.\n");
	am->dwTotalFrames = newFrameCount;
	am->dwMicroSecPerFrame = int(1000000.0/fps + 0.5);
	if (!fwrite(am,sizeof(AVIMAINHEADER),1,o))
	{
		printf("Error writing to output file (0).\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 20;
	}

	/* Next should be an "strl" list for the stream. */
	LISTHEADER strl;
	if (!scanForNextList(f, o, &strl, "strl", true, 0, fmovi))
	{
		printf("strl LIST is invalid or was not found.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 21;
	}

	/* Next should be the "strh" chunk which is an AVISTREAMHEADER structure. */
	if (!readChunk(f, o, "strh", sizeof(AVISTREAMHEADER), &strl))
	{
		printf("Error reading strh CHUNK.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 22;
	}

	/* Make sure that the stream is a video stream. */
	if (((AVISTREAMHEADER*)strl.chkHd->ckData)->fccType != MAKE_FOURCC("vids"))
	{
		printf("The stream present in the file is not a video stream.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 23;
	}

	/* Make sure that AVISF_VIDEO_PALCHANGES isn't set. */
	if (((AVISTREAMHEADER*)strl.chkHd->ckData)->dwFlags&0x00010000)
	{
		printf("Cannot use a stream with AVISF_VIDEO_PALCHANGES flag set.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 24;
	}

	/* Modify needed fields in AVISTREAMHEADER structure. */
	AVISTREAMHEADER* ash = (AVISTREAMHEADER*)strl.chkHd->ckData;
	if (!frac) findRateAndScale(fps, ash->dwRate, ash->dwScale);
	else
	{
		ash->dwRate = num;
		ash->dwScale = den;
	}
	ash->dwLength = newFrameCount;
	if (!fwrite(ash,sizeof(AVISTREAMHEADER),1,o))
	{
		printf("Error writing to output file (1).\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 25;
	}

	/* Next grab the AVI2 "odml" LIST. */
	LISTHEADER odml;
	if (!scanForNextList(f, o, &odml, "odml", false, 0, fmovi))
	{
		if (fmovi) { goto noodml; }
		printf("odml LIST was invalid or was not found.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 26;
	}

	/* Now should be the "dmlh" CHUNK. */
	if (!readChunk(f, o, "dmlh", sizeof(AVIEXTHEADER), &odml))
	{
		printf("Error reading dmlh CHUNK.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 27;
	}

	/* Make sure frame counts match. */
	if (frameCount != ((AVIEXTHEADER*)odml.chkHd->ckData)->dwGrandFrames)
	{
		printf("Frame counts don't match.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 28;
	}

	/* Modify needed fields of AVIEXTHEADER. */
	AVIEXTHEADER *aeh = (AVIEXTHEADER*)odml.chkHd->ckData;
	aeh->dwGrandFrames = newFrameCount;
	if (!fwrite(aeh,sizeof(AVIEXTHEADER),1,o))
	{
		printf("Error writing to output file (2).\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 29;
	}

	/* Next scan to the "movi" list. */
noodml:
	LISTHEADER movi;
	if (!scanForNextList(f, o, &movi, "movi", false, (newFrameCount-frameCount)*8, fmovi))
	{
		printf("movi LIST was invalid or was not found.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 30;
	}

	/* Read all data chunks in the stream and insert drop frames as needed. */
	DWORD i = 0, sum = 0;
	CHUNKHEADER chnk;
	while (i < frameCount)
	{
		if (!readNextChunk(f, o, &chnk, "00dc", true, 0)) break;
		if (!writeEmptyFrames(o, repCount[i], &chnk)) break;
		if (!(i&15))
			printf("\rFrame Modification Progress: %3.2f%c", 
				double(i)*100.0/double(frameCount), '%'); 
		++i;
	}
	printf("\rFrame Modification Progress: %3.2f%c\n", 
		double(i)*100.0/double(frameCount), '%'); 
	if (i != frameCount)
	{
		printf("Number of data chunks does not match frame count.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 31;
	}

	/* Read the "idx1" index and modify the entries as required. */
	i = 0, sum = 0;
	CHUNKHEADER idx1;
	AVIOLDINDEX index;
	if (!readNextChunk(f, o, &idx1, "idx1", false, (newFrameCount-frameCount)*16))
	{
		printf("Error reading idx1 CHUNK.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 32;
	}
	while(1)
	{
		if (i*16 >= idx1.ckSize) break;
		if (!readNextIndexChunk(f, &index)) break;
		if (index.dwChunkId != MAKE_FOURCC("00dc") && index.dwChunkId != MAKE_FOURCC("00db")) 
		{
			if (!fwrite(&index,sizeof(AVIOLDINDEX),1,o)) break;
			break;
		}
		index.dwOffset += sum*8;
		if (!fwrite(&index,sizeof(AVIOLDINDEX),1,o)) break;
		if (!writeNewIndexEntries(o, repCount[i], &index, 
			index.dwOffset+index.dwSize+(index.dwSize&1)+8)) break;
		sum += repCount[i];
		if (!(i&15))
			printf("\rIndex Modification Progress: %3.2f%c", 
				double(i)*100.0/double(frameCount), '%'); 
		++i;
	}
	printf("\rIndex Modification Progress: %3.2f%c\n", 
		double(i)*100.0/double(frameCount), '%');
	if (i != frameCount)
	{
		printf("Number of index chunks does not match frame count.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 33;
	}

	/* flush the end bits */
	while (1)
	{
		DWORD ret = (DWORD)fread(tbuf.data,1,tbuf.size,f);
		if (!ret) break;
		fwrite(tbuf.data,1,ret,o);
	}

	/* Modify the size value of the RIFF Header at the
		beginning of the file. */
	if (fseek(o,0,SEEK_SET))
	{
		printf("Error seeking to beginning of output file.\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 34;
	}
	rh.fileSize += (newFrameCount-frameCount)*24;
	if (!fwrite(&rh,1,12,o))
	{
		printf("Error writing to output file (3).\n");
		closeFiles(f, o, tcf, repCount, true, argv[4], fbuf, obuf);
		return 35;
	}

	/* cleanup and finish */
	finish = clock();
	printf("All processing completed successfully.\n");
	printf("Time elapsed:  %3.3f seconds.\n", (double)(finish-start)/CLOCKS_PER_SEC);
	closeFiles(f, o, tcf, repCount, false, NULL, fbuf, obuf);
	return 0;
}