/*
**                   tcConv v1.0 - July 26, 2006
**
**   This program does v1<->v2 timecode file conversions.
**
**   Copyright (C) 2006 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <stdio.h>
#include <math.h>

#define VERSION "v1.0"

int convertV1toV2(FILE *inf, FILE *outf, int nf, double fpsa);
int convertV2toV1(FILE *inf, FILE *outf, double fpsa, double md);
void closeFiles(FILE *inf, FILE *outf, bool ul, char *name);

int main(int argc, char *argv[])
{
	FILE *tcf = NULL;
	FILE *outf = NULL;

	printf("\ntcConv %s by tritical.\n", VERSION);

	if (argc < 4 || argc > 5)
	{
		printf("Invalid syntax used.\n\n");
		printf("v2 to v1 Usage:  tcConv tc_file_in tc_file_out assume_fps [max_diff]\n");
		printf("Example:  tcConv c:\\timecodes2.txt c:\\timecodes1.txt 29.970\n");
		printf("Example:  tcConv c:\\timecodes2.txt c:\\timecodes1.txt 29.970 0.1\n");
		printf("Example:  tcConv c:\\timecodes2.txt c:\\timecodes1.txt 30000/1001 0.1\n\n");
		printf("v1 to v2 Usage:  tcConv tc_file_in tc_file_out num_frames\n");
		printf("Example:  tcConv c:\\timecodes1.txt c:\\timecodes2.txt 32095\n");
		return 1;
	}

	outf = fopen(argv[2],"w");
	if (!outf)
	{
		printf("Error creating new timecode file.\n");
		return 2;
	}

	tcf = fopen(argv[1],"r");
	if (!tcf)
	{
		printf("Error opening timecode file.\n");
		closeFiles(tcf,outf,true,argv[2]);
		return 3;
	}

	char buf[512];
	if (!fgets(buf,80,tcf)) 
	{
		printf("Unable to read from timecode file.\n");
		closeFiles(tcf,outf,true,argv[2]);
		return 4;
	}
	if (strncmp(buf,"# timecode format v1", 20))
	{
		if (strncmp(buf,"# timecode format v2", 20))
		{
			printf("Invalid timecode file (not v1 or v2).\n");
			closeFiles(tcf,outf,true,argv[2]);
			return 5;
		}
		printf("v2 timecode file type detected.\n");
		double fpsa, md = 0.1;
		char *p = argv[3];
		while (*p != 0 && *p != '/') p++;
		if (*p == 0) 
		{
			fpsa = atof(argv[3]);
			if (fpsa <= 0.0)
			{
				printf("Invalid assume_fps setting (double).\n");
				closeFiles(tcf,outf,true,argv[2]);
				return 6;
			}
		}
		else
		{
			unsigned int num, den;
			if (sscanf(argv[3],"%u/%u", &num, &den) != 2)
			{
				printf("Invalid assume_fps setting (fraction).\n");
				closeFiles(tcf,outf,true,argv[2]);
				return 7;
			}
			if (num == 0 || den == 0)
			{
				printf("Invalid assume_fps setting (fraction).\n");
				closeFiles(tcf,outf,true,argv[2]);
				return 8;
			}
			fpsa = double(num)/double(den);
		}
		if (argc == 5)
		{
			md = atof(argv[4]);
			if (md <= 0.0)
			{
				printf("Invalid max_diff setting.\n");
				closeFiles(tcf,outf,true,argv[2]);
				return 9;
			}
		}
		int error = convertV2toV1(tcf,outf,fpsa,md);
		if (error != 0)
		{
			if (error == 1)
				printf("Error writing to output file.\n");
			else if (error == 2)
				printf("Invalid line detected in tc file.\n");
			else if (error == 3)
				printf("A duration of <= 0.0 was detected.\n");
			else
				printf("An unknown error occurred during processing.\n");
			closeFiles(tcf,outf,true,argv[2]);
			return 10;
		}
	}
	else
	{
		if (!fgets(buf,80,tcf))
		{
			printf("Error reading from timcode file.\n");
			closeFiles(tcf,outf,true,argv[2]);
			return 11;
		}
		printf("v1 timecode file type detected.\n");
		double fpsa = 0.0;
		if (argc != 4)
		{
			printf("Invalid number of arguments for v1->v2 conversion!\n");
			closeFiles(tcf,outf,true,argv[2]);
			return 12;
		}
		if (sscanf(buf,"Assume %lf", &fpsa) != 1)
		{
			printf("Unable to get assumed fps value from v1 file!\n");
			closeFiles(tcf,outf,true,argv[2]);
			return 13;
		}
		if (fpsa <= 0.0)
		{
			printf("Invalid assumed fps value from v1 tc file!\n");
			closeFiles(tcf,outf,true,argv[2]);
			return 14;
		}
		int nf = atoi(argv[3]);
		if (nf <= 0)
		{
			printf("Invalid number of frames for v1->v2 conversion!\n");
			closeFiles(tcf,outf,true,argv[2]);
			return 15;
		}
		int error = convertV1toV2(tcf,outf,nf,fpsa);
		if (error != 0)
		{
			if (error == 1)
				printf("Error writing to output file.\n");
			else if (error == 2 || error == 3)
				printf("Invalid line detected in tc file.\n");
			else
				printf("An unknown error occurred during processing.\n");
			closeFiles(tcf,outf,true,argv[2]);
			return 16;
		}
	}
	printf("All processing completed successfully.\n");
	closeFiles(tcf,outf,false,argv[2]);
	return 0;
}

int convertV2toV1(FILE *inf, FILE *outf, double fpsa, double md)
{
	bool ff = true;
	int sf = 0, cf = -1;
	double st, stf, st_last = 0.0;
	double fps, fps_last = -20.0;
	char line[1025];
	if (fprintf(outf, "# timecode format v1\n") <= 0)
		return 1;
	if (fprintf(outf, "Assume %4.6f\n", fpsa) <= 0)
		return 1;
	while (fgets(line, 1024, inf) != NULL)
	{
		if (line[0] == 0 || line[0] == '\n' || line[0] == '\r' || line[0] == '#')
			continue;
		if (sscanf(line, "%lf", &st) != 1)
			continue;
		if (st < 0.0)
			return 2;
		if (ff)
		{
			stf = st_last = st;
			ff = false;
			++cf;
			continue;
		}
		const double fd = st-st_last;
		if (fd <= 0.0)
			return 3;
		fps = 1000.0/fd;
		if (fps_last == -20.0)
			fps_last = fps;
		if (fabs(fps-fps_last) > md)
		{
			const double afps = (cf-sf)*1000.0/(st_last-stf);
			if (fabs(afps-fpsa) > md)
			{
				if (fprintf(outf, "%d,%d,%4.6f\n", sf, cf-1, afps) <= 0)
					return 1;
			}
			sf = cf;
			stf = st_last;
			fps_last = fps;
		}
		st_last = st;
		++cf;
	}
	const double afps = (cf-sf)*1000.0/(st_last-stf);
	if (fabs(afps-fpsa) > md)
	{
		if (fprintf(outf, "%d,%d,%4.6f\n", sf, cf, afps) <= 0)
			return 1;
	}
	return 0;
}

int convertV1toV2(FILE *inf, FILE *outf, int nf, double fpsa)
{
	double ct = 0.0;
	double mspfa = 1000.0/fpsa;
	int frmStart, frmStop, frmLast = 0;
	char line[1025], *p;
	if (fprintf(outf, "# timecode format v2\n") <= 0)
		return 1;
	while ((fgets(line, 1024, inf) != NULL) && frmLast < nf)
	{
		if (line[0] == 0 || line[0] == '\n' || line[0] == '\r' || line[0] == '#')
			continue;
		p = line;
		while(*p != ',' && *p != 0) p++;
		if (*p == 0) continue;
		p++;
		while(*p != ',' && *p != 0) p++;
		if (*p == 0) continue;
		p++;
		const double fps = atof(p);
		if (fps <= 0.0) continue;
		const double mspf = 1000.0/fps;
		if (sscanf(line, "%d,%d", &frmStart, &frmStop) != 2)
			return 2;
		if (frmStart > frmStop)
			return 3;
		while (frmLast < frmStart && frmLast < nf)
		{
			if (fprintf(outf, "%3.6f\n", ct) <= 0)
				return 1;
			++frmLast;
			ct += mspfa;
		}
		while (frmLast <= frmStop && frmLast < nf)
		{
			if (fprintf(outf, "%3.6f\n", ct) <= 0)
				return 1;
			++frmLast;
			ct += mspf;
		}
	}
	while (frmLast < nf)
	{
		if (fprintf(outf, "%3.6f\n", ct) <= 0)
			return 1;
		++frmLast;
		ct += mspfa;
	}
	return 0;
}

void closeFiles(FILE *inf, FILE *outf, bool ul, char *name)
{
	if (inf) fclose(inf);
	if (outf) fclose(outf);
	if (ul) unlink(name);
}